import json
import queue

from core.debug.debugging import log


class PeerEvents:
    def __init__(self, disconnect):
        self.disconnect = disconnect

        self.messages = queue.Queue()
        self.token_requests = queue.Queue()

    def on_event_received(self, addr, event):
        if not event or not "type" in event:
            log(f"[{addr}] event missing type! {json.dumps(event)}")
            return

        match event["type"]:
            case "message":
                self.messages.put([addr, event])
            case "trusted_token":
                self.token_requests.put([addr, event])
            case "disconnect":
                self.disconnect("disconnect gracefully")
                return
            case _:
                self.disconnect("message type not matched")



from core.storage.storage_manager import StorageManager
from core.models.user_model import UserModel


class UserManager:
    def __init__(self, storage_manager: StorageManager):
        self.storage_manager = storage_manager
        self.contacts = {}

        self.discovered = [] # updated by discovery manager

        self.reload_contacts()

    def set_discovered(self, discovered):
        self.discovered = discovered

    def reload_contacts(self):
        self.storage_manager.ensure_file_exists('contacts')
        self.contacts = {}
        file_content = self.storage_manager.read_file('contacts')
        for contact in file_content:
            self.contacts[contact.verify_key] = UserModel.serialize(contact)

    def get_user(self, verify_key):
        self.ensure_user_exists(verify_key)
        return self.contacts.get(verify_key)

    def ensure_user_exists(self, verify_key, addr=None):
        if verify_key in self.contacts.keys(): return
        self.contacts[verify_key] = UserModel(None, None, verify_key, addr)

    def on_user_discovered(self, verify_key, new_addr):
        self.ensure_user_exists(verify_key, new_addr)
        user = self.contacts.get(verify_key)
        user.addr = new_addr
        self.contacts[verify_key] = user
        if not verify_key in self.discovered:
            self.discovered.append(verify_key)



import inspect


from core.globals import debug


def log(*messages, level=""):
    if not debug: return

    frame = inspect.currentframe().f_back
    cls_name = None

    if 'self' in frame.f_locals:

        cls_name = type(frame.f_locals['self']).__name__

    prefix = f"[{cls_name}]" if cls_name else "no class?"

    message_str = ' '.join(str(m) for m in messages)
    print(f"{level}{prefix}: {message_str}")



import json
from pathlib import Path


class StorageManager:
    def __init__(self, storage_directory=".notepasser"):
        self.storage_path = Path.home() / storage_directory
        self.storage_path.mkdir(parents=True, exist_ok=True)

        self.files = {
            "contacts": self.storage_path / "contacts.json",
            "credentials": self.storage_path / "credentials.json"
        }

        for file_path in self.files.keys():
            self.ensure_file_exists(file_path)

    def ensure_file_exists(self, file_id):
        file_path = self.files.get(file_id)
        if not file_path.exists() or not file_path.is_file():
            file_path.write_text("{}")

    # FILE OPERATIONS
    def read_file(self, file_id):
        with open(self.files.get(file_id), 'r') as f:
            return json.load(f)

    def write_file(self, file_id, new_content):
        with open(self.files.get(file_id), 'w') as f:
            json.dump(new_content, f)




class UserModel:
    version = 1

    def __init__(self, username, bio, verify_key, addr, trusted_token=None):
        self.username = username
        self.bio = bio
        self.verify_key = verify_key
        self.addr = addr
        self.trusted_token = trusted_token

    @classmethod
    def deserialize(cls, file_content):
        return cls(
            username=file_content.get('username'),
            bio=file_content.get('bio'),
            verify_key=file_content.get('verify_key'),
            addr=file_content.get('addr'),
            trusted_token=file_content.get('trusted_token')
        )

    def serialize(self):
        return {
            "version": self.version,
            "username": self.username,
            "bio": self.bio,
            "verify_key": self.verify_key,
            "addr": self.addr,
            "trusted_token": self.trusted_token
        }



import socket
import threading
import time
import traceback

from core.storage.credentials_manager import CredentialsManager
from core.globals import running
from core.networking.peer import Peer
from core.storage.user_manager import UserManager
from core.debug.debugging import log


class NetworkManager:
    def __init__(self, ip, port, credentials_manager: CredentialsManager, user_manager: UserManager, get_trusted_token_input):
        self.credentials_manager = credentials_manager
        self.user_manager = user_manager
        self.get_trusted_token_input = get_trusted_token_input
        self.peers = {}
        self.peers_lock = threading.Lock()

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.bind((ip, port))
        self.sock.listen(5)

        threading.Thread(target=self.listen_for_peers, daemon=True).start()

    def listen_for_peers(self):
        while running:
            try:
                conn, addr = self.sock.accept()
                log("accepted peer " + str(addr))
                if addr in self.peers:
                    log("already exists")
                    continue
                self.peers[addr] = Peer(self, self.user_manager, conn, addr, self.credentials_manager.get_signing_key(), self.get_trusted_token_input)
            except (ConnectionError, socket.timeout):
                continue

    def connect_to_peer(self, verify_key):
        user = self.user_manager.get_user(verify_key)
        peer_ip, peer_port = user.addr
        log("trying to connect to " + str(user.addr))

        # see if peer connection already established
        for peer in self.peers.values():
            timeout = 5
            start = time.time()

            while peer.peer_state.peer_information is None and time.time() - start < timeout:
                time.sleep(0.1)

            log(bytes(peer.peer_state.peer_information['verify_key']).hex(), bytes(verify_key).hex())
            if peer.peer_state.peer_information and bytes(peer.peer_state.peer_information['verify_key']).hex() == bytes(verify_key).hex():
                return peer

        # establish a new peer connection
        try:
            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            conn.connect((peer_ip, peer_port))

            self.peers[(peer_ip, peer_port)] = Peer(self, self.user_manager, conn, (peer_ip, peer_port), self.credentials_manager.get_signing_key(), self.get_trusted_token_input)
            log("connected to " + str((peer_ip, peer_port)))
            return self.peers[(peer_ip, peer_port)]
        except Exception as e:
            log(f"error connecting to {peer_ip}:{peer_port}: {e}", traceback.format_exc())

    def remove_peer(self, peer, reason=None):
        peer.disconnect(reason)
        with self.peers_lock:
            keys_to_remove = [k for k, v in self.peers.items() if v is peer]
            for k in keys_to_remove:
                del self.peers[k]


from nacl.signing import SigningKey

from core.storage.storage_manager import StorageManager


class CredentialsManager:
    def __init__(self, storage_manager: StorageManager):
        self.storage_manager = storage_manager
        self.ensure_credentials_exist()

    def ensure_credentials_exist(self):
        if "signing_key" in self.storage_manager.read_file("credentials").keys(): return

        payload = {
            "signing_key": bytes(SigningKey.generate()).hex()
        }
        self.storage_manager.write_file("credentials", payload)

    def get_signing_key(self):
        self.ensure_credentials_exist()
        return SigningKey(bytes.fromhex(self.storage_manager.read_file("credentials").get("signing_key")))



import json
import queue
import random
import threading
import time
from json import JSONDecodeError
from socket import socket, timeout

from nacl.exceptions import BadSignatureError
from nacl.public import PrivateKey, Box, PublicKey
from nacl.signing import SigningKey, VerifyKey

import core
from core.globals import running
from core.networking.peer_connection import PeerConnection
from core.networking.peer_crypto import PeerCrypto
from core.networking.peer_events import PeerEvents
from core.networking.peer_state import PeerState
from core.storage.user_manager import UserManager
from core.debug.debugging import log


class Peer:
    def __init__(self, network_manager, user_manager: UserManager, conn, addr, my_sign: SigningKey, get_trusted_token):
        self.network_manager = network_manager
        self.user_manager = user_manager
        self.conn = conn
        self.addr = addr

        self.peer_state = PeerState(user_manager, my_sign)
        self.peer_crypto = PeerCrypto(self.peer_state, get_trusted_token)
        self.peer_connection = PeerConnection(self.disconnect, self.peer_crypto, self.conn)
        self.peer_events = PeerEvents(self.disconnect)

        self._disconnected = False
        log("new peer created with random debug number of: " + str(random.randrange(1000, 9999)))
        self.start()

    def start(self):
        threading.Thread(target=self.listen_for_connection_information, daemon=True).start()

        self.conn.sendall(self._get_connection_string())

    def _get_connection_string(self):
        payload = {
            "type": "connection".encode().hex(),
            "encryption_key": bytes(self.peer_crypto.my_pk).hex(),
            "signature": self.peer_crypto.my_signing_key.sign(bytes(self.peer_crypto.my_pk)).signature.hex(),
            "verify_key": bytes(self.peer_crypto.my_verify_key).hex(),
            "trusted_token_exists": bool(self.peer_state.my_information.get("trusted_token"))
        }

        return json.dumps(payload).encode()

    def listen_for_connection_information(self):
        while running and not self._disconnected and not self.peer_state.peer_information.get("verify_key"):
            data = self.peer_connection.receive()

            peer_encryption_key = bytes.fromhex(data["encryption_key"])
            peer_signature = bytes.fromhex(data["signature"])
            peer_verify_key = VerifyKey(bytes.fromhex(data["verify_key"]))
            peer_trusted_token_exists = bool(data["trusted_token_exists"])

            peer_verify_key.verify(peer_encryption_key, peer_signature)

            peer_pk = PublicKey(peer_encryption_key)
            self.peer_crypto.set_my_box(peer_pk)
            self.peer_state.update_peer(peer_verify_key, peer_pk, peer_trusted_token_exists)

            self.peer_state.connected = True

            self.peer_state.reload_user_information()
            self.peer_state.resolve_trusted_state()

        log(self.peer_state.peer_information.get("verify_key"))

        threading.Thread(target=self.listen_for_events, daemon=True).start()

    def listen_for_events(self):
        log("listening for events")
        log(self.conn)

        while running and not self._disconnected:
            event = self.peer_connection.receive()
            log(event)
            self.peer_events.on_event_received(self.addr, event)

        self.disconnect("listen_for_events: no longer running")

    def disconnect(self, reason=None):
        log("disconnecting for : " + str(reason))

        if self._disconnected:
            return
        self._disconnected = True

        # send disconnect only if socket is still valid
        try:
            if self.conn and isinstance(self.conn, socket):
                self.peer_connection.send_disconnect()
        except OSError:
            pass

        try:
            if self.conn and isinstance(self.conn, socket):
                self.conn.close()
        except OSError:
            pass

        self.network_manager.remove_peer(self, reason)



import time

from core.storage.credentials_manager import CredentialsManager
from core.networking.network_manager import NetworkManager
from core.storage.storage_manager import StorageManager


def test_send_and_receive():
    nm1 = NetworkManager("127.0.0.1", 32313, CredentialsManager(StorageManager()))
    nm2 = NetworkManager("127.0.0.1", 32314, CredentialsManager(StorageManager()))

    time.sleep(0.5)

    nm1.connect_to_peer(("127.0.0.1", 32314))
    peer1 = nm1.peers[("127.0.0.1", 32314)]
    while not peer1.box:
        time.sleep(0.1)

    peer2 = list(nm2.peers.values())[0]

    peer1.send_message("hello1")
    peer2.send_message("hello2")

    time.sleep(0.5)

    print("------")
    print(peer1.message_queue.get())
    print(peer1.message_queue.get())
    print("------")
    print(peer2.message_queue.get())
    print(peer2.message_queue.get())


import json
import queue
import time

from nacl.public import PrivateKey, Box
from nacl.signing import SigningKey, VerifyKey

from core.debug.debugging import log
from core.networking.peer_state import PeerState


class PeerCrypto:
    def __init__(self, state: PeerState, get_trusted_token):
        self.state = state

        self.my_signing_key = self.state.my_signing_key
        self.my_verify_key = self.my_signing_key.verify_key
        self.my_sk = PrivateKey.generate()
        self.my_pk = self.my_sk.public_key
        self.box = None

        self.get_trusted_token = get_trusted_token

    def ensure_my_box(self):
        if not self.box: raise RuntimeError("Encryption box is not initialized")

    def encrypt_string(self, message):
        self.ensure_my_box()
        return self.box.encrypt(message.encode())

    def encrypt_json(self, json_message):
        self.ensure_my_box()
        return self.encrypt_string(json.dumps(json_message))

    def decrypt_string(self, encrypted_message):
        self.ensure_my_box()
        return self.box.decrypt(encrypted_message).decode()

    def decrypt_json(self, encrypted_message):
        self.ensure_my_box()
        return json.loads(self.decrypt_string(encrypted_message))

    def set_my_box(self, peer_public_key):
        self.box = Box(self.my_sk, peer_public_key)
